// Author: Pete Kirkham
// tiny string match benchmark.
//
// read binary, match using  benchmark

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <iostream>

#define BLOCK_SIZE (16*1024)
#define MiB (1024*1024)

#include "utils/string_utils.h"

void kmp_table (const char*const string, size_t length, int* table)
{
    size_t  i = 2; // (the current position we are computing in table)
    int     j = 0; // (the zero-based index in string of the next character of the current candidate substring)

//    (the first few values are fixed but different from what the algorithm might suggest)
    table[0] = -1;
    table[1] = 0;

    for (; i < length; )
    {
  //      (first case: the substring continues)
        if (string[i - 1] == string[j])
        {
            table[i] = j + 1;
            ++i;
            ++j;
        }
        else if (j > 0)
        {
            j = table[j];
        }
        else
        {
            table[i] = 0;
            ++i;
        }
    }
}

size_t kmp_search (const char* pattern, const int* table, size_t pattern_length, const char* string, size_t string_length)
{
 /*   input:
        an array of characters, S (the text to be searched)
        an array of characters, W (the word sought)
    output:
        an integer (the zero-based position in S at which W is found)

    define variables:
        an integer, m = 0 (the beginning of the current match in S)
        an integer, i = 0 (the position of the current character in W)
        an array of integers, T (the table, computed elsewhere)
 */
    for (size_t m(0), i(0); (m+1) < string_length; )
    {
        if (pattern[i] == string[m + i])
        {
            i = i + 1;

            if (i == pattern_length)
            {
                return m;
            }
        }
        else
        {
            m = m + i - table[i];

            if (i > 0)
            {
                i = table[i];
            }
        }
    }

    return string_length;
}

size_t matches(0);

size_t scan_line (const char* pattern, const int* table, size_t pattern_length, const char*const line, size_t line_length)
{
    size_t match_pos(kmp_search(pattern, table, pattern_length, line, line_length));

    if (match_pos < line_length)
    {
        ++matches;
//        std::cout << line << "\n";
    }

    return match_pos;
}

int main (int narg, char** argvp)
{
    if (narg < 2)
    {
        std::cout << "The name of the input file is required." << std::endl;
        return 1;
    }

    ssize_t max_read(std::numeric_limits<ssize_t>::max());

    if (narg > 2)
    {
        max_read = atoi(argvp[2]) * MiB;
    }

    // open the file
    int file(open(argvp[1], O_RDONLY));

    if (!file)
    {
        std::cout << "failed to open " << argvp[1] << std::endl;
        return 2;
    }

    // find length
    struct stat file_stat;

    if (fstat(file, &file_stat))
    {
        std::cout << "failed to stat " << argvp[1] << std::endl;
        return 2;
    }

    // generate KMP table
    const char* pattern("GET /ongoing/When/");
    size_t      pattern_length(strlen(pattern));
    int*        table(new int[pattern_length]);

    kmp_table(pattern, pattern_length, table);

    // read in file

    size_t      file_length(file_stat.st_size);

#ifdef BLOCK_READ

    char        buf0[BLOCK_SIZE];
    char        buf1[BLOCK_SIZE];
    char*       buf(buf0);
    ssize_t     read_now;
    ssize_t     read_total(0);
    ssize_t     MiBMask((ssize_t(MiB)*8)-1);
    size_t      line_start(0);

    while ((read_now = read(file, buf, BLOCK_SIZE)) > 0)
    {
        read_total += read_now;

        // scan for lines
        for (size_t i(0); i < BLOCK_SIZE; ++i)
        {
            if (buf[i] == '\n')
            {
                if (line_start > i)
                {
                    // handle split over
                    memcpy(buf1 + (BLOCK_SIZE - line_start), buf, i);

                    buf1[i + (BLOCK_SIZE - line_start)] = '\0';

                    scan_line(pattern, table, pattern_length, buf1, i + (BLOCK_SIZE - line_start));
                }
                else
                {
                    buf[i] = '\0';
                    scan_line(pattern, table, pattern_length, buf + line_start, i - line_start);
                }

                line_start = i+1;
            }
        }

        // copy trailing data to start of alternate block
        memcpy(buf1, buf0+line_start, BLOCK_SIZE - line_start);

        if ((read_total & MiBMask) == 0)
        {
            if (read_total >= max_read) break;

            std::cout << "read " << read_total/MiB << " MiB." << std::endl;
        }
    }

    std::cout << "read " << read_total/MiB << " MiB." << std::endl;
#else

    char*       buf(new char[file_length]);
    ssize_t     read_total(0);

    while ((read_total += read(file, buf + read_total, file_length)) < file_length)
        std::cout << "read " << read_total/MiB << " MiB." << std::endl;

#ifdef LINES
    // scan for lines
    for (size_t i(0), line_start(0); i < file_length; ++i)
    {
        if (buf[i] == '\n')
        {
            buf[i] = '\0';
            scan_line(pattern, table, pattern_length, buf + line_start, i - line_start);
            line_start = i+1;
        }
    }
#else
    // scan for matches directly
    for (size_t match_pos(0); match_pos < file_length; ++match_pos)
    {
        match_pos += scan_line(pattern, table, pattern_length, buf + match_pos, file_length - match_pos);
    }
#endif

    delete [] buf;
#endif
    close(file);

    delete [] table;

    std::cout << "matches: " << matches << std::endl;

    return 0;
}
